/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  backend-db31.c defines the DocBook V.3.1 output backend of refdbd
  markus@mhoenicka.de 6-22-00
  $Id: backend-db31.c,v 1.16.2.7 2006/02/11 00:45:54 mhoenicka Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#include <string.h>
#include <syslog.h> /* for definitions of log message priorities */
#include <dbi/dbi.h>

#include "refdb.h"
#include "backend.h"
#include "backend-db31.h"
#include "strfncs.h"
#include "linklist.h"
#include "tokenize.h"
#include "dbfncs.h"
#include "risdb.h"
#include "authorinfo.h"

/* some globals */
extern int n_log_level; /* numeric version of log_level */

/* forward declarations of local functions */
static char* split_author_db(char* author, char* dbauthor);
static char* add_author_db(char* authorstring, struct AUTHOR_INFO* ptr_ainfo);
static char* add_author_info_db(struct renderinfo* ptr_rendinfo, int author_type, int n_refdb_id, char** ptr_chunk_buf, size_t* ptr_chunk_buf_len);
static char* add_analytic_db(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len, const char* type);
static char* add_monogr_db(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len, const char* type);
static char* add_series_db(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len, const char* type);
static char* add_notes_db(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len);
static char* add_abstract_db(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len);

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  prepare_render_db31(): writes a header for the docbook output of a
                         query

  int prepare_render_db31 returns 0 if successful,>0 if failed

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int prepare_render_db31(struct renderinfo* ptr_rendinfo) {
  /* ToDo: control with format_string whether or not to include the document type declaration. For the time being the declaration does go in but is commented out */

  if (ptr_rendinfo->ref_format == REFDOCBK) { /* SGML */
    strcpy(*(ptr_rendinfo->ptr_ref), "<!-- <!DOCTYPE bibliography PUBLIC \"-//OASIS//DTD DocBook V3.1//EN\"> -->\n<bibliography>\n<title>Bibliography</title>\n");
  }
  else if (ptr_rendinfo->ref_format == REFDOCBKX) { /* XML */
    strcpy(*(ptr_rendinfo->ptr_ref), "<!-- <?xml version=\"1.0\"?><!DOCTYPE bibliography PUBLIC \"-//OASIS//DTD DocBook XML V4.1.2//EN\" \"http://www.oasis-open.org/docbook/xml/4.1.2/docbookx.dtd\"> -->\n<bibliography>\n<title>Bibliography</title>\n");
  }
  else {
    /* we have a problem */
    LOG_PRINT(LOG_WARNING, get_status_msg(302));
    return 302;
  }

  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  finish_render_db31(): writes a footer for the docbook output of a query

  int finish_render_ris returns 0 if successful, >0 if failed

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int finish_render_db31(struct renderinfo* ptr_rendinfo) {
  char* new_ref;

  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</bibliography>\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 801;
  }
  
  *(ptr_rendinfo->ptr_ref) = new_ref;

  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  render_db31() renders a RIS dataset for DocBook V.3.1 export

  int render_db31 returns 0 if successful, >0 if not

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int render_db31(struct renderinfo* ptr_rendinfo) {
  char have_author = 0; 
  char id[32] = ""; /* reference ID */
  char *new_ref; /* used for reallocating buffer ref */
  char date_buffer[256]; /* holds date */
  char *dbauthor;
  char *entitize_buf; /* buffer for replacing &<> with sgml entities */
  char *chunk;
  char *chunk_buf;
  const char *item; /* pointer to string for all db requests */
  const char *type;
  int i; /* the tried-and-true counter */
  size_t entitize_buf_len = 0; /* length of entitize_buf */
  size_t chunk_buf_len;
  unsigned long long n_refdb_id;
  dbi_result dbires;
  dbi_conn conn;
  dbi_conn conn_refdb;

  /* rendering bibliographic information takes up to three levels of
     information:
     analytic: author and title of the article or chapter
     monographic: name of the journal/book, vol/issue/page info, editors
     series: series title, series editors
  */

  chunk_buf_len = 1024;

  if ((chunk_buf = malloc(chunk_buf_len)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 801;
  }

  conn = dbi_result_get_conn(ptr_rendinfo->dbires);

  if ((type = get_refdb_type(ptr_rendinfo->dbires)) == NULL) {
    LOG_PRINT(LOG_WARNING, get_status_msg(234));
    free(chunk_buf);
    return 234;
  }

  n_refdb_id = my_dbi_result_get_idval(ptr_rendinfo->dbires, "refdb_id");

  if ((ptr_rendinfo->ptr_biblio_info)->entry_id != NULL) {
    strcat(*(ptr_rendinfo->ptr_ref), "<biblioentry id=\"");
    strcat(*(ptr_rendinfo->ptr_ref), (ptr_rendinfo->ptr_biblio_info)->entry_id);
    strcat(*(ptr_rendinfo->ptr_ref), "\">\n<abbrev>");
  }
  else {
    strcat(*(ptr_rendinfo->ptr_ref), "<biblioentry>\n<abbrev>"); /* ToDo: use databasename? */
  }
  item = get_refdb_citekey(ptr_rendinfo->dbires);
  if (item && *item) {
    strcat(*(ptr_rendinfo->ptr_ref), item); /* ID */
  }

  /* up to here it's gonna fit anyway, but now we rather use mstrcat() */
  
  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</abbrev>\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    free(chunk_buf);
    return 801;
  }
  else {
    *(ptr_rendinfo->ptr_ref) = new_ref;
  }

  chunk = add_analytic_db(ptr_rendinfo, n_refdb_id, &chunk_buf, &chunk_buf_len, type);

  if (chunk && *chunk) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), chunk, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      free(chunk_buf);
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }

  chunk = add_monogr_db(ptr_rendinfo, n_refdb_id, &chunk_buf, &chunk_buf_len, type);

  if (chunk && *chunk) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), chunk, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      free(chunk_buf);
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }

  chunk = add_series_db(ptr_rendinfo, n_refdb_id, &chunk_buf, &chunk_buf_len, type);

  if (chunk && *chunk) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), chunk, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      free(chunk_buf);
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }

  /* add notes and abstract if requested */

  /* notes */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "N1") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    chunk = add_notes_db(ptr_rendinfo, n_refdb_id, &chunk_buf, &chunk_buf_len);

    if (chunk && *chunk) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), chunk, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	free(chunk_buf);
	return 801;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }
  }

  /*----------------------------------------------------------------*/
  /* abstract */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "N2") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "AB") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    chunk = add_abstract_db(ptr_rendinfo, n_refdb_id, &chunk_buf, &chunk_buf_len);

    if (chunk && *chunk) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), chunk, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	free(chunk_buf);
	return 801;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }
  }

  /* finish bibliography entry */
  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</biblioentry>\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    free(chunk_buf);
    return 801;
  }
  else {
    *(ptr_rendinfo->ptr_ref) = new_ref;
  }

  free(chunk_buf);

  return 0;
}
 
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  add_analytic_db() adds analytic level info to a DocBook bibliography

  static char* add_analytic_db returns ptr to ref if successful, NULL if not

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  int n_refdb_id ID of the current dataset

  char **ptr_chunk_buf ptr to an allocated string, may be resized

  size_t *ptr_chunk_buf_len ptr to length of ptr_chunk_buf

  const char* type ptr to string holding the type of the current dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static char* add_analytic_db(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len, const char* type) {
  int n_have_data = 0; /* set to 1 if we have some data */
  char *new_ref;
  char *entitize_buf;
  char *author_buf;
  char *author_result;
  const char *item;
  size_t author_buf_len;

  **ptr_chunk_buf = '\0';

  /* only these types are supposed to have analytic information */
  if (!strcmp(type, "CHAP") || !strcmp(type, "JOUR")
      || !strcmp(type, "MGZN") || !strcmp(type, "NEWS")
      || !strcmp(type, "CONF")) {
    author_buf_len = 1024;

    if ((author_buf = malloc(author_buf_len)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    
    /* start analytic information */
    if ((new_ref = mstrcat(*ptr_chunk_buf, "<biblioset relation=\"analytic\">\n", ptr_chunk_buf_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      free(author_buf);
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
    
    /* add authors, if any */
    author_result = add_author_info_db(ptr_rendinfo, 1, n_refdb_id, &author_buf, &author_buf_len);
    
    if (!author_result) {
      free(author_buf);
      return NULL;
    }
    else if (*author_result) {
      n_have_data++;
      if ((new_ref = mstrcat(*ptr_chunk_buf, author_result, ptr_chunk_buf_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	free(author_buf);
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }
    }
    
    free(author_buf);
    
    /* get the article or chapter title */
    item = get_refdb_title_copy(ptr_rendinfo->dbires);
/*     if (!item) { */
/*       item = get_refdb_booktitle_copy(ptr_rendinfo->dbires); */
/*     } */
    
    if (item != NULL && *item) {
      n_have_data++;
      if ((entitize_buf = mstrdup((char*)item)) == NULL) {
	free((char*)item);
	return NULL;
      }
      
      if (sgml_entitize(&entitize_buf, NULL) == NULL) {
	free((char*)item);
	free(entitize_buf);
	return NULL;
      }
      
      if ((new_ref = mstrcat(*ptr_chunk_buf, "<title>", ptr_chunk_buf_len, 0)) == NULL) {
	free((char*)item);
	free(entitize_buf);
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }
      
      if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* title */
	free((char*)item);
	free(entitize_buf);
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }
      
      if ((new_ref = mstrcat(*ptr_chunk_buf, "</title>\n", ptr_chunk_buf_len, 0)) == NULL) {
	free((char*)item);
	free(entitize_buf);
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }
      free(entitize_buf);
      free((char*)item);
    }

    /* finish analytic information */
    if ((new_ref = mstrcat(*ptr_chunk_buf, "</biblioset>\n", ptr_chunk_buf_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if (!n_have_data) {
      **ptr_chunk_buf = '\0';
    }
  }

  return *ptr_chunk_buf;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  add_monogr_db() adds monography level info to a DocBook bibliography

  static char* add_monogr_db returns ptr to ref if successful, NULL if not

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  int n_refdb_id ID of the current dataset

  char **ptr_chunk_buf ptr to an allocated string, may be resized

  size_t *ptr_chunk_buf_len ptr to length of ptr_chunk_buf

  const char* type ptr to string holding the type of the current dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static char* add_monogr_db(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len, const char* type) {
  int n_have_data = 0; /* set to 1 if we have some data */
  int i;
  int errcode; /* receives error code for periodical requests */
  char periodical[256];
  char *new_ref;
  char *entitize_buf;
  char *author_buf;
  char *author_result;
  const char *item;
  char journaltype[4][12] = {"name",
			     "custabbrev1",
			     "custabbrev2",
			     "abbrev"};
  size_t author_buf_len;
  dbi_conn conn;

  
  **ptr_chunk_buf = '\0';

  author_buf_len = 1024;

  if ((author_buf = malloc(author_buf_len)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
    
  conn = dbi_result_get_conn(ptr_rendinfo->dbires);

  /* start monographic information */
  if ((new_ref = mstrcat(*ptr_chunk_buf, "<biblioset relation=\"monogr\">\n", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    free(author_buf);
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  /* chapters and articles and such may have editors here. All others
     should have authors here */

  if (!strcmp(type, "CHAP") || !strcmp(type, "JOUR")
      || !strcmp(type, "MGZN") || !strcmp(type, "NEWS")
      || !strcmp(type, "CONF")) {
    /* add editors, if any */
    author_result = add_author_info_db(ptr_rendinfo, 2, n_refdb_id, &author_buf, &author_buf_len);

    if (!author_result) {
      free(author_buf);
      return NULL;
    }
    else if (*author_result) {
      n_have_data++;
      if ((new_ref = mstrcat(*ptr_chunk_buf, author_result, ptr_chunk_buf_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	free(author_buf);
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }
    }
  }
  else {
    /* add authors, if any */
    author_result = add_author_info_db(ptr_rendinfo, 1, n_refdb_id, &author_buf, &author_buf_len);

    if (!author_result) {
      free(author_buf);
      return NULL;
    }
    else if (*author_result) {
      n_have_data++;
      if ((new_ref = mstrcat(*ptr_chunk_buf, author_result, ptr_chunk_buf_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	free(author_buf);
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }
    }
  }

  free(author_buf);

  if (!strcmp(type, "JFULL") || !strcmp(type, "JOUR")
      || !strcmp(type, "MGZN") || !strcmp(type, "NEWS")
      || !strcmp(type, "GEN")) {
    /* journal information, if any */
    for (i = 0; i < 4; i++) {
      /* request journal */
      item = get_periodical(conn, periodical, /*ptr_rendinfo->database*/NULL, i, &errcode, 0, my_dbi_result_get_idval(ptr_rendinfo->dbires, "refdb_id"), NULL /* no frequency required */);
      if (item != NULL) {
	if (!*item) {
	  continue;
	}
	if ((entitize_buf = mstrdup((char*)item)) == NULL) {
	  return NULL;
	}
    
	if (sgml_entitize(&entitize_buf, NULL) == NULL) {
	  free(entitize_buf);
	  return NULL;
	}
	
	if ((new_ref = mstrcat(*ptr_chunk_buf, "<biblioset relation=\"", ptr_chunk_buf_len, 0)) == NULL) {
	  free(entitize_buf);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return NULL;
	}
	else {
	  *ptr_chunk_buf = new_ref;
	}
	
	if ((new_ref = mstrcat(*ptr_chunk_buf, journaltype[i], ptr_chunk_buf_len, 0)) == NULL) {
	  free(entitize_buf);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return NULL;
	}
	else {
	  *ptr_chunk_buf = new_ref;
	}

	if ((new_ref = mstrcat(*ptr_chunk_buf, "\">\n", ptr_chunk_buf_len, 0)) == NULL) {
	  free(entitize_buf);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return NULL;
	}
	else {
	  *ptr_chunk_buf = new_ref;
	}
	
	if ((new_ref = mstrcat(*ptr_chunk_buf, "<title>", ptr_chunk_buf_len, 0)) == NULL) {
	  free(entitize_buf);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return NULL;
	}
	else {
	  *ptr_chunk_buf = new_ref;
	}

	if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) {
	  free(entitize_buf);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return NULL;
	}
	else {
	  *ptr_chunk_buf = new_ref;
	}

	if ((new_ref = mstrcat(*ptr_chunk_buf, "</title>\n</biblioset>\n", ptr_chunk_buf_len, 0)) == NULL) {
	  free(entitize_buf);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return NULL;
	}
	else {
	  *ptr_chunk_buf = new_ref;
	}
	free(entitize_buf);
	n_have_data++;
      }
    }
  }

  /* book title or primary title */
  item = NULL;
  if (!strcmp(type, "BOOK")/*  || !strcmp(type, "UNPB")  */
      || !strcmp(type, "CHAP") || !strcmp(type, "CONF")
      || !strcmp(type, "GEN")) {
    item = get_refdb_booktitle_copy(ptr_rendinfo->dbires);
  }
  else if (strcmp(type, "JFULL") && strcmp(type, "JOUR")
      && strcmp(type, "MGZN")  && strcmp(type, "NEWS")
      && strcmp(type, "GEN")) {
    item = get_refdb_title_copy(ptr_rendinfo->dbires);
  }

  if (item != NULL) {
    if ((entitize_buf = mstrdup((char*)item)) == NULL) {
      free((char*)item);
      return NULL;
    }

    if (sgml_entitize(&entitize_buf, NULL) == NULL) {
      free((char*)item);
      free(entitize_buf);
      return NULL;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "<biblioset relation=\"booktitle\"><title>", ptr_chunk_buf_len, 0)) == NULL) {
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* title */
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "</title></biblioset>\n", ptr_chunk_buf_len, 0)) == NULL) {
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
    free(entitize_buf);
    free((char*)item);
    n_have_data++;
  }

  /* ISBN/ISSN */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "SN") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    if ((item = get_refdb_issn(ptr_rendinfo->dbires)) != NULL) {
      if ((new_ref = mstrcat(*ptr_chunk_buf, "<isbn>", ptr_chunk_buf_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }

      if ((new_ref = mstrcat(*ptr_chunk_buf, (char*)item, ptr_chunk_buf_len, 0)) == NULL) { /* isbn/issn */
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }
      
      if ((new_ref = mstrcat(*ptr_chunk_buf, "</isbn>\n", ptr_chunk_buf_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }
      n_have_data++;
    }
  }

  /* publisher and city */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "PB") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    if ((item = get_refdb_publisher(ptr_rendinfo->dbires)) != NULL) {
      if ((new_ref = mstrcat(*ptr_chunk_buf, "<publisher><publishername>", ptr_chunk_buf_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }

      if ((entitize_buf = mstrdup((char*)item)) == NULL) {
	return NULL;
      }

      if (sgml_entitize(&entitize_buf, NULL) == NULL) {
	free(entitize_buf);
	return NULL;
      }

      if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* publisher */
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }
      
      free(entitize_buf);
      
      if ((new_ref = mstrcat(*ptr_chunk_buf, "</publishername>", ptr_chunk_buf_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }
      
      if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "CY") != NULL
	  || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
	/* pubplace - use only if publisher is known */
	if ((item = get_refdb_city(ptr_rendinfo->dbires)) != NULL) {
	  if ((new_ref = mstrcat(*ptr_chunk_buf, "<address>", ptr_chunk_buf_len, 0)) == NULL) {
	    LOG_PRINT(LOG_CRIT, get_status_msg(801));
	    return NULL;
	  }
	  else {
	    *ptr_chunk_buf = new_ref;
	  }

	  if ((new_ref = mstrcat(*ptr_chunk_buf, (char*)item, ptr_chunk_buf_len, 0)) == NULL) { /* city */
	    LOG_PRINT(LOG_CRIT, get_status_msg(801));
	    return NULL;
	  }
	  else {
	    *ptr_chunk_buf = new_ref;
	  }

	  if ((new_ref = mstrcat(*ptr_chunk_buf, "</address>", ptr_chunk_buf_len, 0)) == NULL) {
	    LOG_PRINT(LOG_CRIT, get_status_msg(801));
	    return NULL;
	  }
	  else {
	    *ptr_chunk_buf = new_ref;
	  }
	}
      }
      
      if ((new_ref = mstrcat(*ptr_chunk_buf, "</publisher>\n", ptr_chunk_buf_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }

      n_have_data++;
    }
  }

  if ((item = get_refdb_volume(ptr_rendinfo->dbires)) != NULL) {
    if ((entitize_buf = mstrdup((char*)item)) == NULL) {
      return NULL;
    }

    if (sgml_entitize(&entitize_buf, NULL) == NULL) {
      free(entitize_buf);
      return NULL;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "<volumenum>", ptr_chunk_buf_len, 0)) == NULL) {
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* volume */
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "</volumenum>\n", ptr_chunk_buf_len, 0)) == NULL) {
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
    free(entitize_buf);
    n_have_data++;
  }

  if ((item = get_refdb_issue(ptr_rendinfo->dbires)) != NULL) {
    if ((entitize_buf = mstrdup((char*)item)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }

    if (sgml_entitize(&entitize_buf, NULL) == NULL) {
      free(entitize_buf);
      return NULL;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "<issuenum>", ptr_chunk_buf_len, 0)) == NULL) {
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* issue */
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "</issuenum>\n", ptr_chunk_buf_len, 0)) == NULL) {
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
    free(entitize_buf);
    n_have_data++;
  }

  if ((item = get_refdb_startpage(ptr_rendinfo->dbires)) != NULL) {
    if ((entitize_buf = mstrdup((char*)item)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }

    if (sgml_entitize(&entitize_buf, NULL) == NULL) {
      free(entitize_buf);
      return NULL;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "<pagenums>", ptr_chunk_buf_len, 0)) == NULL) {
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* start page */
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
    free(entitize_buf);
    n_have_data++;

    if ((item = get_refdb_endpage(ptr_rendinfo->dbires)) != NULL) {
      if ((entitize_buf = mstrdup((char*)item)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }

      if (sgml_entitize(&entitize_buf, NULL) == NULL) {
	free(entitize_buf);
	return NULL;
      }

      if ((new_ref = mstrcat(*ptr_chunk_buf, "-", ptr_chunk_buf_len, 0)) == NULL) {
	free(entitize_buf);
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }

      if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* end page */
	free(entitize_buf);
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }
      free(entitize_buf);
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "</pagenums>\n", ptr_chunk_buf_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
  }

  /* reuse periodical */
  if ((item = get_refdb_pubyear(ptr_rendinfo->dbires, periodical)) != NULL) {
    if ((new_ref = mstrcat(*ptr_chunk_buf, "<pubdate>", ptr_chunk_buf_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, periodical, ptr_chunk_buf_len, 0)) == NULL) { /* year */
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "</pubdate>\n", ptr_chunk_buf_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
    n_have_data++;
  }

  /* finish monographic information */
  if ((new_ref = mstrcat(*ptr_chunk_buf, "</biblioset>\n", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  if (!n_have_data) {
    **ptr_chunk_buf = '\0';
  }

  return *ptr_chunk_buf;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  add_series_db() adds series level info to a DocBook bibliography

  static char* add_series_db returns ptr to ref if successful, NULL if not

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  int n_refdb_id ID of the current dataset

  char **ptr_chunk_buf ptr to an allocated string, may be resized

  size_t *ptr_chunk_buf_len ptr to length of ptr_chunk_buf

  const char* type ptr to string holding the type of the current dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static char* add_series_db(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len, const char* type) {
  int n_have_data = 0; /* set to 1 if we have some data */
  char *new_ref;
  char *entitize_buf;
  char *author_buf;
  char *author_result;
  const char *item;
  size_t author_buf_len;

  **ptr_chunk_buf = '\0';

  author_buf_len = 1024;

  if ((author_buf = malloc(author_buf_len)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
    
  /* start series info */
  if ((new_ref = mstrcat(*ptr_chunk_buf, "<biblioset relation=\"series\">\n", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    free(author_buf);
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  /* add editors, if any */
  author_result = add_author_info_db(ptr_rendinfo, 3, n_refdb_id, &author_buf, &author_buf_len);

  if (!author_result) {
    free(author_buf);
    return NULL;
  }
  else if (*author_result) {
    n_have_data++;
    if ((new_ref = mstrcat(*ptr_chunk_buf, author_result, ptr_chunk_buf_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      free(author_buf);
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
  }

  free(author_buf);

  /* series title, if any */
  item = get_refdb_title_series_copy(ptr_rendinfo->dbires);

  if (item && *item) {
    n_have_data++;
    if ((entitize_buf = mstrdup((char*)item)) == NULL) {
      free((char*)item);
      return NULL;
    }

    if (sgml_entitize(&entitize_buf, NULL) == NULL) {
      free((char*)item);
      free(entitize_buf);
      return NULL;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "<title>", ptr_chunk_buf_len, 0)) == NULL) {
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* title */
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "</title>\n", ptr_chunk_buf_len, 0)) == NULL) {
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
    free(entitize_buf);
    free((char*)item);
    n_have_data++;
  }

  /* finish series information */
  if ((new_ref = mstrcat(*ptr_chunk_buf, "</biblioset>\n", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  if (!n_have_data) {
    **ptr_chunk_buf = '\0';
  }

  return *ptr_chunk_buf;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  add_notes_db() adds notes info to a DocBook bibliography

  static char* add_notes_db returns ptr to ref if successful, NULL if not

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  int n_refdb_id ID of the current dataset

  char **ptr_chunk_buf ptr to an allocated string, may be resized

  size_t *ptr_chunk_buf_len ptr to length of ptr_chunk_buf

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static char* add_notes_db(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len) {
  int n_have_data = 0; /* set to 1 if we have some data */
  char *new_ref;
  char *entitize_buf;
  const char *item;

  **ptr_chunk_buf = '\0';

  /* start bibliomisc element */
  if ((new_ref = mstrcat(*ptr_chunk_buf, "<bibliomisc role=\"annotation\">", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  /* note of current user, if any */
  item = get_notes_copy(ptr_rendinfo->dbires, ptr_rendinfo->username);

  if (item && *item) {
    n_have_data++;
    if ((entitize_buf = mstrdup((char*)item)) == NULL) {
      free((char*)item);
      return NULL;
    }

    if (sgml_entitize(&entitize_buf, NULL) == NULL) {
      free((char*)item);
      free(entitize_buf);
      return NULL;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* title */
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
  }

  /* finish notes */
  if ((new_ref = mstrcat(*ptr_chunk_buf, "</bibliomisc>\n", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  if (!n_have_data) {
    **ptr_chunk_buf = '\0';
  }

  return *ptr_chunk_buf;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  add_abstract_db() adds abstract to a DocBook bibliography

  static char* add_abstract_db returns ptr to ref if successful, NULL if not

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  int n_refdb_id ID of the current dataset

  char **ptr_chunk_buf ptr to an allocated string, may be resized

  size_t *ptr_chunk_buf_len ptr to length of ptr_chunk_buf

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static char* add_abstract_db(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len) {
  int n_have_data = 0; /* set to 1 if we have some data */
  char *new_ref;
  char *entitize_buf;
  const char *item;

  **ptr_chunk_buf = '\0';

  /* start abstract */
  if ((new_ref = mstrcat(*ptr_chunk_buf, "<abstract>\n", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  /* abstract, if any */
  item = get_refdb_abstract_copy(ptr_rendinfo->dbires);

  if (item != NULL) {
    n_have_data++;
    if ((entitize_buf = mstrdup((char*)item)) == NULL) {
      free((char*)item);
      return NULL;
    }

    if (sgml_entitize(&entitize_buf, NULL) == NULL) {
      free((char*)item);
      free(entitize_buf);
      return NULL;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "<para>", ptr_chunk_buf_len, 0)) == NULL) {
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* abstract */
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "</para>\n", ptr_chunk_buf_len, 0)) == NULL) {
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
    free(entitize_buf);
    free((char*)item);
  }

  /* finish abstract information */
  if ((new_ref = mstrcat(*ptr_chunk_buf, "</abstract>\n", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  if (!n_have_data) {
    **ptr_chunk_buf = '\0';
  }

  return *ptr_chunk_buf;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  split_author_db(): converts a RIS author into a sequence of DocBook
                  elements

  static char* split_author_db returns a pointer to dbauthor

  char* dbauthor ptr to a buffer which receives the DocBook result

  char* author ptr to a buffer with RIS author

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static char* split_author_db(char* dbauthor, char* author) {
  struct AUTHORTOKENS atokens;
  Lilistring* ptr_middle;

  *dbauthor = '\0'; /* start with empty string */

  if (!tokenize_author(author, &atokens)) {
    return dbauthor;
  }

  /* assemble string */
  if (*atokens.first) {
    strcat(dbauthor, "<firstname>");
    strcat(dbauthor, atokens.first);
    strcat(dbauthor, "</firstname>");
  }

  if (atokens.ptr_middlelist) {
    ptr_middle = atokens.ptr_middlelist;
    while ((ptr_middle = get_next_lilistring(ptr_middle)) != NULL) {
      strcat(dbauthor, "<othername role=\"mi\">");
      strcat(dbauthor, ptr_middle->token);
      strcat(dbauthor, "</othername>");
    }
  }

  if (*atokens.sur) {
    strcat(dbauthor, "<surname>");
    strcat(dbauthor, atokens.sur);
    strcat(dbauthor, "</surname>");
  }

  if (*atokens.lineage) {
    strcat(dbauthor, "<lineage>");
    strcat(dbauthor, atokens.lineage);
    strcat(dbauthor, "</lineage>");
  }

  clean_authortokens(&atokens);
  return dbauthor;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  add_author_db(): adds an author in DocBook format

  static char* add_author_db returns a pointer to authorstring

  char* authorstring ptr to a buffer which receives the DocBook result

  struct AUTHOR_INFO* ptr_ainfo ptr to a structure with the name parts

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static char* add_author_db(char* authorstring, struct AUTHOR_INFO* ptr_ainfo) {
  char* item;
  char* entitize_buf;

  *authorstring = '\0'; /* start with empty string */

  if (!ptr_ainfo) {
    return authorstring;
  }

  /* ToDo: check for corporate name */
  if (!*(ptr_ainfo->lastname) && !*(ptr_ainfo->firstname)
      && !*(ptr_ainfo->middlename) && !*(ptr_ainfo->suffix)) {
    if (*(ptr_ainfo->role)) {
      strcat(authorstring, "<corpauthor role=\"");
      strcat(authorstring, ptr_ainfo->role);
      strcat(authorstring, "\">");
    }
    else {
      strcat(authorstring, "<corpauthor>");
    }

    if ((entitize_buf = mstrdup(ptr_ainfo->name)) == NULL) {
      return authorstring;
    }

    if (sgml_entitize(&entitize_buf, NULL) == NULL) {
      free(entitize_buf);
      return authorstring;
    }
    strcat(authorstring, "</corpauthor>");
  }
  else { /* have name parts */
    if (*(ptr_ainfo->role)) {
      strcat(authorstring, "<author role=\"");
      strcat(authorstring, ptr_ainfo->role);
      strcat(authorstring, "\">");
    }
    else {
      strcat(authorstring, "<author>");
    }

    if (*(ptr_ainfo->firstname)) {
      strcat(authorstring, "<firstname>");
      if ((entitize_buf = mstrdup(ptr_ainfo->firstname)) == NULL) {
	return authorstring;
      }

      if (sgml_entitize(&entitize_buf, NULL) == NULL) {
	free(entitize_buf);
	return authorstring;
      }

      strcat(authorstring, entitize_buf);
      free(entitize_buf);
      strcat(authorstring, "</firstname>");
    }

    if (*(ptr_ainfo->middlename)) {
      for (item = strtok(ptr_ainfo->middlename, " "); item; item = strtok(NULL, " ")) {
	strcat(authorstring, "<othername role=\"mi\">");
	if ((entitize_buf = mstrdup(item)) == NULL) {
	  return authorstring;
	}

	if (sgml_entitize(&entitize_buf, NULL) == NULL) {
	  free(entitize_buf);
	  return authorstring;
	}

	strcat(authorstring, entitize_buf);
	free(entitize_buf);
	strcat(authorstring, "</othername>");
      }
    }

    if (*(ptr_ainfo->lastname)) {
      strcat(authorstring, "<surname>");
      if ((entitize_buf = mstrdup(ptr_ainfo->lastname)) == NULL) {
	return authorstring;
      }

      if (sgml_entitize(&entitize_buf, NULL) == NULL) {
	free(entitize_buf);
	return authorstring;
      }

      strcat(authorstring, entitize_buf);
      free(entitize_buf);
      strcat(authorstring, "</surname>");
    }

    if (*(ptr_ainfo->suffix)) {
      strcat(authorstring, "<lineage>");
      if ((entitize_buf = mstrdup(ptr_ainfo->suffix)) == NULL) {
	return authorstring;
      }

      if (sgml_entitize(&entitize_buf, NULL) == NULL) {
	free(entitize_buf);
	return authorstring;
      }

      strcat(authorstring, entitize_buf);
      free(entitize_buf);
      strcat(authorstring, "</lineage>");
    }
    strcat(authorstring, "</author>");
  }

  return authorstring;
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  add_author_info_db() adds author info to a DocBook bibliography

  static char* add_author_info_db returns ptr to ref if successful, NULL if not

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  int author_type 1=author 2=editor 3=series editor

  int n_refdb_id ID of the current dataset

  char** ptr_chunk_buf ptr to a buffer to receive output

  size_t* ptr_chunk_buf_len ptr to length of string

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static char* add_author_info_db(struct renderinfo* ptr_rendinfo, int author_type, int n_refdb_id, char** ptr_chunk_buf, size_t* ptr_chunk_buf_len) {
  int nis_first = 1;
  char *new_ref;
  char *dbauthor;
  const char *item;
  dbi_conn conn;
  dbi_result dbires;
  struct AUTHOR_INFO ainfo;

  **ptr_chunk_buf = '\0';

  conn = dbi_result_get_conn(ptr_rendinfo->dbires);

  /* get authors */
  dbires = request_authors(conn, author_type, NULL, 0, n_refdb_id);
  if (dbires == NULL) {
    return NULL;
  }

  /* get buffer to assemble the docbook author output */
  dbauthor = malloc((size_t)2048);  /* fixed length ok afaik */
    
  if (dbauthor == NULL) {
    clean_request(dbires);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }

  /* fetch all authors of this article */
  while (get_author_parts(dbires, &ainfo) != NULL) {
    /* begin authorgroup element */
    if (nis_first) {
      nis_first--;
      if ((new_ref = mstrcat(*ptr_chunk_buf, "<authorgroup>\n", ptr_chunk_buf_len, 0)) == NULL) {
	clean_request(dbires);
	free(dbauthor);
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, add_author_db(dbauthor, &ainfo), ptr_chunk_buf_len, 0)) == NULL) {
      clean_request(dbires);
      free(dbauthor);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
  }
  free(dbauthor);
  clean_request(dbires);

  /* close authorgroup element only if it was started previously */
  if (!nis_first) {
    if ((new_ref = mstrcat(*ptr_chunk_buf, "\n</authorgroup>\n", ptr_chunk_buf_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
  }
  else {
    **ptr_chunk_buf = '\0';
  }

  /* if there was no author info, chunk_buf is returned unaltered */
  return *ptr_chunk_buf;
}
