/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  backend-scrn.c defines the screen output backend of refdbd
  markus@mhoenicka.de 6-20-00
  $Id: backend-scrn.c,v 1.17.2.6 2005/11/03 20:40:19 mhoenicka Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#include <string.h>
#include <syslog.h> /* for definitions of log message priorities */
#include <dbi/dbi.h>

#include "refdb.h"
#include "linklist.h"
#include "refdbd.h"
#include "backend.h"
#include "backend-scrn.h"
#include "writenote.h"
#include "strfncs.h"
#include "dbfncs.h"
#include "connect.h"

#ifndef HAVE_ATOLL
long long atoll(const char *str);
#endif

/* some globals */
extern int n_log_level; /* numeric version of log_level */

/* forward declaration of local functions */
static char* print_field_scrn(const char* item, struct renderinfo* ptr_rendinfo, const char* start_string);

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  prepare_render_scrn(): writes a header for the screen output of a
                         query

  int prepare_render_scrn returns 0 if successful, >0 if failed

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int prepare_render_scrn(struct renderinfo* ptr_rendinfo) {
  /* we just make sure that we start with a clean string */
  (*(ptr_rendinfo->ptr_ref))[0] = '\0';
  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  finish_render_scrn(): writes a footer for the screen output of a query

  int finish_render_scrn returns 0 if successful, >0 if failed

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int finish_render_scrn(struct renderinfo* ptr_rendinfo) {
  /* nothing to do here */
  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  render_scrn() renders a RIS dataset for screen display

  int render_scrn returns 0 if successful, >0 if failed

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int render_scrn(struct renderinfo* ptr_rendinfo) {
  int errcode; /* receives error code for periodical requests */
  int nhave_booktitle = 0;
  int render_res;
  char have_author = 0;
  char id[32] = "";
  char avail_type[6];
  char date_buffer[256];
  char* avail;
  char* item;
  char* new_ref;
  const char* citem;
  const char* citem1;
  const char* type;
  struct REPRINT reprint;
  dbi_result dbires;
  dbi_result orig_dbires;
  dbi_conn conn;

  if ((type = get_refdb_type(ptr_rendinfo->dbires)) == NULL) {
    LOG_PRINT(LOG_WARNING, get_status_msg(234));
    return 234;
  }

  conn = dbi_result_get_conn(ptr_rendinfo->dbires);

  /*----------------------------------------------------------------*/
  /* ID */
  get_refdb_id(ptr_rendinfo->dbires, id);

  if (is_in_list(ptr_rendinfo->dbires, ptr_rendinfo->username) == 1) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "ID*:", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }
  else {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "ID:", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }

  if (*id) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), id, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }
  else { /* reference has no ID */
    LOG_PRINT(LOG_WARNING, get_status_msg(234));
    return 234; /* this is bad and will hopefully never happen */
  }

  /*----------------------------------------------------------------*/
  /* pubyear */
  if (get_refdb_pubyear(ptr_rendinfo->dbires, date_buffer) != NULL) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), " (", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    /* year */
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), date_buffer, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), ")", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }

  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 801;
  }
  else {
    *(ptr_rendinfo->ptr_ref) = new_ref;
  }

  /*----------------------------------------------------------------*/
  /* citation key */
  citem = get_refdb_citekey(ptr_rendinfo->dbires);
  if (citem && *citem) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "Key: ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)citem, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }

  /*----------------------------------------------------------------*/
  /* authors */
  /* request authors */
  dbires = request_authors(conn, 1, NULL, 0, my_dbi_result_get_idval(ptr_rendinfo->dbires, "refdb_id"));
  if (dbires == NULL) {
    LOG_PRINT(LOG_WARNING, get_status_msg(234));
    return 234;
  }

  /* fetch all authors of this article */
  while ((citem = get_author(dbires)) != NULL) {
    have_author = 1;
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)citem, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      clean_request(dbires);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), ", ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      clean_request(dbires);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }
  clean_request(dbires);

  if (have_author) { /* eliminate the trailing comma after the last author */
    new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\n", ptr_rendinfo->ptr_ref_len, 2);
  }
  else {
    new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\n", ptr_rendinfo->ptr_ref_len, 0);
  }

  if (new_ref == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 801;
  }
  else {
    *(ptr_rendinfo->ptr_ref) = new_ref;
  }

  /*----------------------------------------------------------------*/
  /* title */
  item = get_refdb_title_copy(ptr_rendinfo->dbires);
  if (item != NULL) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* title */
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      free(item);
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

    free(item);

    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      free(item);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }

  if (strcmp(type, "BOOK") == 0 || strcmp(type, "CHAP") == 0
      || strcmp(type, "SER") == 0 || strcmp(type, "THES") == 0
      || strcmp(type, "RPRT") == 0 || strcmp(type, "CONF") == 0
      || strcmp(type, "UNPB") == 0) {
    if (strcmp(type, "CHAP") == 0) {
      /*----------------------------------------------------------------*/
      /* book editors */
      dbires = request_authors(conn, 2, NULL, 0, my_dbi_result_get_idval(ptr_rendinfo->dbires, "refdb_id")); /* book ed. */
      if (dbires == NULL) {
	return 234;
      }

      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\nin: ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return 801;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }

      have_author = 0;
      /* fetch all editors of this article */
      while ((citem = get_author(dbires)) != NULL) {
	have_author = 1;
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)citem, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  clean_request(dbires);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return 801;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), ", ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  clean_request(dbires);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return 801;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
      }
      clean_request(dbires);

      if (have_author) { /* eliminate the trailing comma after the last author */
	new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "(ed.), ", ptr_rendinfo->ptr_ref_len, 2);
      }
      
      if (new_ref == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return 801;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }

      /*----------------------------------------------------------------*/
      /* book title */
      item = get_refdb_booktitle_copy(ptr_rendinfo->dbires);
      if (item != NULL) {
	nhave_booktitle++;
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* book title */
	  free(item);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return 801;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
	free(item);
      }
    }
    else { /* not chapter */
      /*----------------------------------------------------------------*/
      /* book title */
      item = get_refdb_booktitle_copy(ptr_rendinfo->dbires);
      if (item != NULL) {
	nhave_booktitle++;
/* 	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) { */
/* 	  free(item); */
/* 	  LOG_PRINT(LOG_CRIT, get_status_msg(801)); */
/* 	  return NULL; */
/* 	} */
/* 	else { */
/* 	  *(ptr_rendinfo->ptr_ref) = new_ref; */
/* 	} */
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* book title */
	  free(item);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return 801;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
	free(item);
      }
    }

    item = get_refdb_title_series_copy(ptr_rendinfo->dbires);
    if (item != NULL) {
      dbires = request_authors(conn, 3, NULL, 0, my_dbi_result_get_idval(ptr_rendinfo->dbires, "refdb_id")); /* series ed. */
      if (dbires == NULL) {
	free(item);
	return 234;
      }
	
      if (nhave_booktitle) {
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "; part of: ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  free(item);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  clean_request(dbires);
	  return 801;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
      }

      have_author = 0;
      /* fetch all editors of this article */
      while ((citem = get_author(dbires)) != NULL) {
	have_author = 1;
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)citem, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  free(item);
	  clean_request(dbires);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return 801;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), ", ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  free(item);
	  clean_request(dbires);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return 801;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
      }
      clean_request(dbires);
      
      if (have_author) { /* eliminate the trailing comma after the last author */
	new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "(series ed.), ", ptr_rendinfo->ptr_ref_len, 2);
      }
      
      if (new_ref == NULL) {
	free(item);
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return 801;
      }
      
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* series title */
	free(item);
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return 801;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
      free(item);
    }
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }
  else {
  /*----------------------------------------------------------------*/
  /* periodical */
    citem = get_periodical(dbi_result_get_conn(ptr_rendinfo->dbires), date_buffer, NULL, 4, &errcode, 0, my_dbi_result_get_idval(ptr_rendinfo->dbires, "refdb_id"), NULL /* no frequency required */);
    if (citem == NULL && errcode != 1) {
      return 234;
    }
    else if (!errcode) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)citem, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return 801;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }

    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), " ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }

  }

  /*----------------------------------------------------------------*/
  /* volume */
  citem = get_refdb_volume(ptr_rendinfo->dbires);
  if (citem != NULL) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)citem, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* volume */
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }
  
  /*----------------------------------------------------------------*/
  /* issue */
  citem = get_refdb_issue(ptr_rendinfo->dbires);  
  if (citem != NULL) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "(", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)citem, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* issue */
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), ")", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }

  /*----------------------------------------------------------------*/
  /* startpage */
  citem = get_refdb_startpage(ptr_rendinfo->dbires);
  if (citem != NULL) {
    if (strcmp(type, "BOOK") == 0 || strcmp(type, "SER") == 0 ||
	strcmp(type, "THES") == 0 || strcmp(type, "RPRT") == 0 ||
	strcmp(type, "CHAP") == 0) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "pp. ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return 801;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }
    else {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), ":", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return 801;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }

    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)citem, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* start page */
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }
  
  /*----------------------------------------------------------------*/
  /* endpage */
  citem = get_refdb_endpage(ptr_rendinfo->dbires);
  if (citem != NULL) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "-", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)citem, ptr_rendinfo->ptr_ref_len, 0)) == NULL) { /* end page */
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }
  
  /*----------------------------------------------------------------*/
  /* abstract */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "N2") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "AB") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    if ((item = get_refdb_abstract_copy(ptr_rendinfo->dbires)) != NULL) {
      if (print_field_scrn(item, ptr_rendinfo, "\nABSTRACT: ") == NULL) {
	free(item);
	return 801;
      }
      free(item);
    }
  }

  /*----------------------------------------------------------------*/
  /* address */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "AD") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    if ((item = get_refdb_address_copy(ptr_rendinfo->dbires)) != NULL) {
      if (print_field_scrn(item, ptr_rendinfo, "\nADDRESS: ") == NULL) {
	free(item);
	return 801;
      }
      free(item);
    }
  }

  /*----------------------------------------------------------------*/
  /* reprint, availability  */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "RP") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    if (get_reprint(ptr_rendinfo->dbires, &reprint, ptr_rendinfo->username, 4) != NULL
	&&reprint.reprint != NULL && *(reprint.reprint)) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\nREPRINT: ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return 801;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), reprint.reprint, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return 801;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
      if (reprint.date != NULL && *(reprint.date)) {
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), " (", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return 801;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), reprint.date, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return 801;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), ")", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return 801;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
      }
    }
    if (reprint.avail != NULL && *(reprint.avail)) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), ": ", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return 801;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }

      /* copy the relevant part of reprint.avail so we can safely
	 uppercase it */
      strncpy(avail_type, reprint.avail, 5);
      avail_type[5] = '\0';

      if (strncmp(strup(avail_type), "PATH:", 5) == 0) {
	if (*(ptr_rendinfo->pdfroot)) {
	  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), ptr_rendinfo->pdfroot, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	    LOG_PRINT(LOG_CRIT, get_status_msg(801));
	    return 801;
	  }
	  else {
	    *(ptr_rendinfo->ptr_ref) = new_ref;
	  }
	}
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), reprint.avail+5, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return 801;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
      }
      else {
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), reprint.avail, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return 801;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}
      }
    }
  }

  /*----------------------------------------------------------------*/
  /* ISBN/ISSN */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "SN") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    citem = get_refdb_issn(ptr_rendinfo->dbires);
    if (print_field_scrn(citem, ptr_rendinfo, "\nISSN: ") == NULL) {
      return 801;
    }
  }

  /*----------------------------------------------------------------*/
  /* URL, L1 through L4 */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "UR") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    int i;
    char link_tag[12];

    /* loop over all link types */
    for (i=0; i<5;i++) {
      if (!i) {
	strcpy(link_tag, "\nURL: ");
      }
      else if (i == 1) {
	strcpy(link_tag, "\nPDF: ");
      }
      else if (i == 2) {
	strcpy(link_tag, "\nFULLTEXT: ");
      }
      else if (i == 3) {
	strcpy(link_tag, "\nRELATED: ");
      }
      else if (i == 4) {
	strcpy(link_tag, "\nIMAGE: ");
      }
      
      dbires = request_ulinks(conn, my_dbi_result_get_idval(ptr_rendinfo->dbires, "refdb_id"), 0 /* ref entry */, i /* link type */, 0 /* is_temp */);
      if (dbires == NULL) {
	return 234;
      }

      while ((citem = get_ulink(dbires)) != NULL) {
	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), link_tag, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  clean_request(dbires);
	  return 801;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}

	if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)citem, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  clean_request(dbires);
	  return 801;
	}
	else {
	  *(ptr_rendinfo->ptr_ref) = new_ref;
	}

      } /* end while */
      clean_request(dbires);
    } /* end for */
  }

  /*----------------------------------------------------------------*/
  /* publisher */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "PB") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    citem = get_refdb_publisher(ptr_rendinfo->dbires);
    if (print_field_scrn(citem, ptr_rendinfo, "\nPUBLISHER: ") == NULL) {
      return 801;
    }
  }

  /*----------------------------------------------------------------*/
  /* city */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "CY") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    citem = get_refdb_city(ptr_rendinfo->dbires);
    if (print_field_scrn(citem, ptr_rendinfo, "\nCITY: ") == NULL) {
      return 801;
    }
  }

  /*----------------------------------------------------------------*/
  /* user fields */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "U1") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    citem = get_refdb_user1(ptr_rendinfo->dbires);
    if (print_field_scrn(citem, ptr_rendinfo, "\nUSER1: ") == NULL) {
      return 801;
    }
  }

  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "U2") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    citem = get_refdb_user2(ptr_rendinfo->dbires);
    if (print_field_scrn(citem, ptr_rendinfo, "\nUSER2: ") == NULL) {
      return 801;
    }
  }

  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "U3") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    citem = get_refdb_user3(ptr_rendinfo->dbires);
    if (print_field_scrn(citem, ptr_rendinfo, "\nUSER3: ") == NULL) {
      return 801;
    }
  }

  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "U4") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    citem = get_refdb_user4(ptr_rendinfo->dbires);
    if (print_field_scrn(citem, ptr_rendinfo, "\nUSER4: ") == NULL) {
      return 801;
    }
  }

  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "U5") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    citem = get_refdb_user5(ptr_rendinfo->dbires);
    if (print_field_scrn(citem, ptr_rendinfo, "\nUSER5: ") == NULL) {
      return 801;
    }
  }

  /*----------------------------------------------------------------*/
  /* misc fields */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "M1") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    citem = get_refdb_misc1(ptr_rendinfo->dbires);
    if (print_field_scrn(citem, ptr_rendinfo, "\nMISC1: ") == NULL) {
      return 801;
    }
  }

  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "M2") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    citem = get_refdb_misc2(ptr_rendinfo->dbires);
    if (print_field_scrn(citem, ptr_rendinfo, "\nMISC2: ") == NULL) {
      return 801;
    }
  }

  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "M3") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    citem = get_refdb_misc3(ptr_rendinfo->dbires);
    if (print_field_scrn(citem, ptr_rendinfo, "\nMISC3: ") == NULL) {
      return 801;
    }
  }


  /*----------------------------------------------------------------*/
  /* the notes */

  /* first the note which is attached to the reference */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "N1") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    if ((item = get_notes_copy(ptr_rendinfo->dbires, ptr_rendinfo->username)) != NULL) {
      if (print_field_scrn(item, ptr_rendinfo, "\nNOTES: ") == NULL) {
	free(item);
	return 801;
      }
      free(item);
    }
  }

  /* then look for additional notes linked with the record,
     keywords, authors, or periodicals */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "NX") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    if ((dbires = request_notes_by_ref(dbi_result_get_conn(ptr_rendinfo->dbires), atoll(id), REFERENCE|KEYWORD|AUTHOR|PERIODICAL, ptr_rendinfo->username, (ptr_rendinfo->ptr_clrequest)->share_default)) == NULL) {
/*       printf("request_notes_by_ref() failed\n"); */
      free_request_notes_by_ref(dbi_result_get_conn(ptr_rendinfo->dbires));
      return 234;
    }
    
    /* save original query result. This is legal as dbi_result is a
       pointer in disguise */
    orig_dbires = ptr_rendinfo->dbires;
    ptr_rendinfo->dbires = dbires;

    while (dbi_result_next_row(ptr_rendinfo->dbires) != 0) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\nXNOTE:\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	free_request_notes_by_ref(dbi_result_get_conn(ptr_rendinfo->dbires));
	return 801;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }

      if ((render_res = render_note(ptr_rendinfo)) != 0) {
	ptr_rendinfo->dbires = orig_dbires;
	clean_request(dbires);
	free_request_notes_by_ref(dbi_result_get_conn(ptr_rendinfo->dbires));
	return render_res;
      }
    }
    clean_request(dbires);
    free_request_notes_by_ref(dbi_result_get_conn(orig_dbires));
    ptr_rendinfo->dbires = orig_dbires;
  }

  /* finish with an empty line */
  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "\n\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 801;
  }
  else {
    *(ptr_rendinfo->ptr_ref) = new_ref;
  }
/*      printf("%s\n", ref); */
  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  print_field_scrn(): writes simple field contents to HTML output

  static char* print_field_scrn returns a ptr to the modified string. Due to
                         reallocation this may be different from the
			 ptr passed to this function in the ptr_rendinfo
			 structure which is also updated accordingly

  const char* item ptr to the contents of the field to write

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  const char* start_string ptr to a string to be printed before the 
                         field contents

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static char* print_field_scrn(const char* item, struct renderinfo* ptr_rendinfo, const char* start_string) {
  char* new_ref;

  if (item != NULL) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)start_string, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), (char*)item, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }
  return *(ptr_rendinfo->ptr_ref);
}
