/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  backend-teix.c defines the TEI XML output backend of refdbd
  markus@mhoenicka.de 2002-08-27
  $Id: backend-teix.c,v 1.7.2.4 2005/11/03 20:40:19 mhoenicka Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#include <string.h>
#include <syslog.h> /* for definitions of log message priorities */
#include <dbi/dbi.h>

#include "refdb.h"
#include "backend.h"
#include "backend-teix.h"
#include "strfncs.h"
#include "dbfncs.h"
#include "linklist.h"
#include "risdb.h"
#include "authorinfo.h"

/* some globals */
extern int n_log_level; /* numeric version of log_level */

/* forward declarations of local functions */
static char* add_author_info_teix(struct renderinfo* ptr_rendinfo, int author_type, int n_refdb_id, char** ptr_chunk_buf, size_t* ptr_chunk_buf_len);
static char* add_analytic_teix(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len, const char* type);
static char* add_monogr_teix(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len, const char* type);
static char* add_series_teix(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len, const char* type);
static char* add_notes_teix(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len);
static char* add_abstract_teix(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len);

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  prepare_render_teix(): writes a header for the TEI XML output of a
                         query

  int prepare_render_teix returns 0 if successful, >0 if failed

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int prepare_render_teix(struct renderinfo* ptr_rendinfo) {
  /* ToDo: control with format_string whether or not to include the document type declaration. For the time being the declaration does go in but is commented out */

  strcpy(*(ptr_rendinfo->ptr_ref), "<!-- <?xml version=\"1.0\"?>\n<!DOCTYPE listBibl PUBLIC \"-//TEI P4//DTD Main Document Type//EN\" \"http://www.tei-c.org/P4X/DTD/tei2.dtd\" [\n<!ENTITY % TEI.general \'INCLUDE\'>\n<!ENTITY % TEI.names.dates \'INCLUDE\'>\n<!ENTITY % TEI.linking \'INCLUDE\'>\n<!ENTITY % TEI.XML \'INCLUDE\'>\n]> -->\n<listBibl>\n<head>Bibliography</head>\n");
  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  finish_render_teix(): writes a footer for the TEI XML output of a query

  int finish_render_ris returns 0 if successful, >0 if failed

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int finish_render_teix(struct renderinfo* ptr_rendinfo) {
  char* new_ref;

  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</listBibl>\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 801;
  }
  *(ptr_rendinfo->ptr_ref) = new_ref;

  return 0;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  render_teix() renders a RIS dataset for TEI XML export

  int render_teix returns 0 if successful, >0 if not

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int render_teix(struct renderinfo* ptr_rendinfo) {
  char have_author = 0; 
  char id[32] = ""; /* reference ID */
  char *new_ref; /* used for reallocating buffer ref */
  char date_buffer[256]; /* holds date */
  char *dbauthor;
  char *entitize_buf; /* buffer for replacing &<> with sgml entities */
  char *chunk;
  char *chunk_buf;
  const char *item; /* pointer to string for all db requests */
  const char *type;
  int i; /* the tried-and-true counter */
  size_t entitize_buf_len = 0; /* length of entitize_buf */
  size_t chunk_buf_len;
  unsigned long long n_refdb_id;
  dbi_result dbires;
  dbi_conn conn;
  dbi_conn conn_refdb;

  /* rendering bibliographic information takes up to three levels of
     information:
     analytic: author and title of the article or chapter
     monographic: name of the journal/book, vol/issue/page info, editors
     series: series title, series editors
  */

  chunk_buf_len = 1024;

  if ((chunk_buf = malloc(chunk_buf_len)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return 801;
  }

  conn = dbi_result_get_conn(ptr_rendinfo->dbires);

  if ((type = get_refdb_type(ptr_rendinfo->dbires)) == NULL) {
    free(chunk_buf);
    LOG_PRINT(LOG_WARNING, get_status_msg(234));
    return 234;
  }

  n_refdb_id = my_dbi_result_get_idval(ptr_rendinfo->dbires, "refdb_id");

  strcat(*(ptr_rendinfo->ptr_ref), "<biblStruct id=\"");
  item = get_refdb_citekey(ptr_rendinfo->dbires);
  if (item && *item) {
    strcat(*(ptr_rendinfo->ptr_ref), item);
  }
  else {
    strcat(*(ptr_rendinfo->ptr_ref), (ptr_rendinfo->ptr_biblio_info)->entry_id);
  }
  strcat(*(ptr_rendinfo->ptr_ref), "\">\n");

  /* up to here it's gonna fit anyway, but now we rather use mstrcat() */
  
  chunk = add_analytic_teix(ptr_rendinfo, n_refdb_id, &chunk_buf, &chunk_buf_len, type);

  if (chunk && *chunk) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), chunk, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      free(chunk_buf);
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }

  chunk = add_monogr_teix(ptr_rendinfo, n_refdb_id, &chunk_buf, &chunk_buf_len, type);

  if (chunk && *chunk) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), chunk, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      free(chunk_buf);
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }

  chunk = add_series_teix(ptr_rendinfo, n_refdb_id, &chunk_buf, &chunk_buf_len, type);

  if (chunk && *chunk) {
    if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), chunk, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      free(chunk_buf);
      return 801;
    }
    else {
      *(ptr_rendinfo->ptr_ref) = new_ref;
    }
  }

  /* add notes and abstract if requested */

  /* notes */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "N1") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    chunk = add_notes_teix(ptr_rendinfo, n_refdb_id, &chunk_buf, &chunk_buf_len);

    if (chunk && *chunk) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), chunk, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	free(chunk_buf);
	return 801;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }
  }

  /*----------------------------------------------------------------*/
  /* abstract */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "N2") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "AB") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    chunk = add_abstract_teix(ptr_rendinfo, n_refdb_id, &chunk_buf, &chunk_buf_len);

    if (chunk && *chunk) {
      if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), chunk, ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	free(chunk_buf);
	return 801;
      }
      else {
	*(ptr_rendinfo->ptr_ref) = new_ref;
      }
    }
  }

  /* finish bibliography entry */
  if ((new_ref = mstrcat(*(ptr_rendinfo->ptr_ref), "</biblStruct>\n", ptr_rendinfo->ptr_ref_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    free(chunk_buf);
    return 801;
  }
  else {
    *(ptr_rendinfo->ptr_ref) = new_ref;
  }

  free(chunk_buf);

  return 0;
}
 
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  add_analytic_teix() adds analytic level info to a TEI XML bibliography

  static char* add_analytic_teix returns ptr to ref if successful, NULL if not

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  int n_refdb_id ID of the current dataset

  char **ptr_chunk_buf ptr to an allocated string, may be resized

  size_t *ptr_chunk_buf_len ptr to length of ptr_chunk_buf

  const char* type ptr to string containing reference type

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static char* add_analytic_teix(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len, const char* type) {
  int n_have_data = 0; /* set to 1 if we have some data */
  char *new_ref;
  char *entitize_buf;
  char *author_buf;
  char *author_result;
  const char *item;
  size_t author_buf_len;

  **ptr_chunk_buf = '\0';

  author_buf_len = 1024;

  if ((author_buf = malloc(author_buf_len)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
    
  /* start analytic information */
  if ((new_ref = mstrcat(*ptr_chunk_buf, "<analytic>\n", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    free(author_buf);
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  /* add authors, if any */
  author_result = add_author_info_teix(ptr_rendinfo, 1, n_refdb_id, &author_buf, &author_buf_len);

  if (!author_result) {
    free(author_buf);
    return NULL;
  }
  else if (*author_result) {
    n_have_data++;
    if ((new_ref = mstrcat(*ptr_chunk_buf, author_result, ptr_chunk_buf_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      free(author_buf);
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
  }

  free(author_buf);

  /* get the article or book title */
  item = get_refdb_title_copy(ptr_rendinfo->dbires);
  if (!item) {
    item = get_refdb_booktitle_copy(ptr_rendinfo->dbires);
  }

  if (item != NULL && *item) {
    n_have_data++;
    if ((entitize_buf = mstrdup((char*)item)) == NULL) {
      free((char*)item);
      return NULL;
    }

    if (sgml_entitize(&entitize_buf, NULL) == NULL) {
      free((char*)item);
      free(entitize_buf);
      return NULL;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "<title level=\"a\">", ptr_chunk_buf_len, 0)) == NULL) {
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* title */
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "</title>\n", ptr_chunk_buf_len, 0)) == NULL) {
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
    free(entitize_buf);
    free((char*)item);
  }

  /* finish analytic information */
  if ((new_ref = mstrcat(*ptr_chunk_buf, "</analytic>\n", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  if (!n_have_data) {
    **ptr_chunk_buf = '\0';
  }

  return *ptr_chunk_buf;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  add_monogr_teix() adds monography level info to a TEI XML bibliography

  static char* add_monogr_teix returns ptr to ref if successful, NULL if not

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  int n_refdb_id ID of the current dataset

  char **ptr_chunk_buf ptr to an allocated string, may be resized

  size_t *ptr_chunk_buf_len ptr to length of ptr_chunk_buf

  const char* type ptr to string containing reference type

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static char* add_monogr_teix(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len, const char* type) {
  int n_have_data = 0; /* set to 1 if we have some data */
  int i;
  int errcode; /* receives error code for periodical requests */
  char periodical[256];
  char *new_ref;
  char *entitize_buf;
  char *author_buf;
  char *author_result;
  const char *item;
  char journaltype[4][12] = {"name",
			     "custabbrev1",
			     "custabbrev2",
			     "abbrev"};
  size_t author_buf_len;
  dbi_conn conn;

  
  **ptr_chunk_buf = '\0';

  author_buf_len = 1024;

  if ((author_buf = malloc(author_buf_len)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
    
  conn = dbi_result_get_conn(ptr_rendinfo->dbires);

  /* start monographic information */
  if ((new_ref = mstrcat(*ptr_chunk_buf, "<monogr>\n", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    free(author_buf);
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  /* add editors, if any */
  author_result = add_author_info_teix(ptr_rendinfo, 2, n_refdb_id, &author_buf, &author_buf_len);

  if (!author_result) {
    free(author_buf);
    return NULL;
  }
  else if (*author_result) {
    n_have_data++;
    if ((new_ref = mstrcat(*ptr_chunk_buf, author_result, ptr_chunk_buf_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      free(author_buf);
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
  }

  free(author_buf);

  if (!strcmp(type, "JFULL") || !strcmp(type, "JOUR")
      || !strcmp(type, "MGZN") || !strcmp(type, "NEWS")
      || !strcmp(type, "GEN")) {
    /* journal information, if any */
    for (i = 0; i < 4; i++) {
      /* request journal */
      item = get_periodical(conn, periodical, /*ptr_rendinfo->database*/NULL, i, &errcode, 0, my_dbi_result_get_idval(ptr_rendinfo->dbires, "refdb_id"), NULL /* no frequency required */);
      if (item != NULL) {
	if (!*item) {
	  continue;
	}
	if ((entitize_buf = mstrdup((char*)item)) == NULL) {
	  return NULL;
	}
    
	if (sgml_entitize(&entitize_buf, NULL) == NULL) {
	  free(entitize_buf);
	  return NULL;
	}
      
/*       if ((new_ref = mstrcat(*ptr_chunk_buf, journaltype[i], ptr_chunk_buf_len, 0)) == NULL) { */
/* 	free(entitize_buf); */
/* 	LOG_PRINT(LOG_CRIT, get_status_msg(801)); */
/* 	return NULL; */
/*       } */
/*       else { */
/* 	*ptr_chunk_buf = new_ref; */
/*       } */

	if ((new_ref = mstrcat(*ptr_chunk_buf, "<title level=\"j\">", ptr_chunk_buf_len, 0)) == NULL) {
	  free(entitize_buf);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return NULL;
	}
	else {
	  *ptr_chunk_buf = new_ref;
	}

	if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) {
	  free(entitize_buf);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return NULL;
	}
	else {
	  *ptr_chunk_buf = new_ref;
	}

	if ((new_ref = mstrcat(*ptr_chunk_buf, "</title>\n", ptr_chunk_buf_len, 0)) == NULL) {
	  free(entitize_buf);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return NULL;
	}
	else {
	  *ptr_chunk_buf = new_ref;
	}
	free(entitize_buf);
	n_have_data++;
      }
    }
  }

  /* book title or primary title */
  item = NULL;
  if (!strcmp(type, "BOOK") || !strcmp(type, "UNPB")
      || !strcmp(type, "CHAP") || !strcmp(type, "CONF")
      || !strcmp(type, "GEN")) {
    item = get_refdb_booktitle_copy(ptr_rendinfo->dbires);
  }
  else if (strcmp(type, "JFULL") && strcmp(type, "JOUR")
      && strcmp(type, "MGZN")  &strcmp(type, "NEWS")
      && strcmp(type, "GEN")) {
    item = get_refdb_title_copy(ptr_rendinfo->dbires);
  }

  if (item != NULL) {
    if ((entitize_buf = mstrdup((char*)item)) == NULL) {
      free((char*)item);
      return NULL;
    }

    if (sgml_entitize(&entitize_buf, NULL) == NULL) {
      free((char*)item);
      free(entitize_buf);
      return NULL;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "<title level=\"m\">", ptr_chunk_buf_len, 0)) == NULL) {
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* title */
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "</title>\n", ptr_chunk_buf_len, 0)) == NULL) {
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
    free(entitize_buf);
    free((char*)item);
    n_have_data++;
  }

  if ((new_ref = mstrcat(*ptr_chunk_buf, "<imprint>\n", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  /* ISBN/ISSN */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "SN") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    if ((item = get_refdb_issn(ptr_rendinfo->dbires)) != NULL) {
      if ((new_ref = mstrcat(*ptr_chunk_buf, "<idno type=\"isbn\">", ptr_chunk_buf_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }

      if ((new_ref = mstrcat(*ptr_chunk_buf, (char*)item, ptr_chunk_buf_len, 0)) == NULL) { /* isbn/issn */
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }

      if ((new_ref = mstrcat(*ptr_chunk_buf, "</isbn>", ptr_chunk_buf_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }
      n_have_data++;
    }
  }

  /* volume/issue info */
  if ((item = get_refdb_volume(ptr_rendinfo->dbires)) != NULL) {
    if ((entitize_buf = mstrdup((char*)item)) == NULL) {
      return NULL;
    }

    if (sgml_entitize(&entitize_buf, NULL) == NULL) {
      free(entitize_buf);
      return NULL;
    }
    
    /* volume */
    if ((new_ref = mstrcat(*ptr_chunk_buf, "<biblScope type=\"vol\">", ptr_chunk_buf_len, 0)) == NULL) {
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* volume */
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "</biblScope>\n", ptr_chunk_buf_len, 0)) == NULL) {
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
    free(entitize_buf);
    n_have_data++;
  }

  /* issue */
  if ((item = get_refdb_issue(ptr_rendinfo->dbires)) != NULL) {
    if ((entitize_buf = mstrdup((char*)item)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }

    if (sgml_entitize(&entitize_buf, NULL) == NULL) {
      free(entitize_buf);
      return NULL;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "<biblScope type=\"issue\">", ptr_chunk_buf_len, 0)) == NULL) {
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* issue */
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "</biblScope>\n", ptr_chunk_buf_len, 0)) == NULL) {
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
    free(entitize_buf);
    n_have_data++;
  }

  /* pages */
  if ((item = get_refdb_startpage(ptr_rendinfo->dbires)) != NULL) {
    if ((entitize_buf = mstrdup((char*)item)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }

    if (sgml_entitize(&entitize_buf, NULL) == NULL) {
      free(entitize_buf);
      return NULL;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "<biblScope type=\"pages\">", ptr_chunk_buf_len, 0)) == NULL) {
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* start page */
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
    free(entitize_buf);
    n_have_data++;

    if ((item = get_refdb_endpage(ptr_rendinfo->dbires)) != NULL) {
      if ((entitize_buf = mstrdup((char*)item)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }

      if (sgml_entitize(&entitize_buf, NULL) == NULL) {
	free(entitize_buf);
	return NULL;
      }

      if ((new_ref = mstrcat(*ptr_chunk_buf, "-", ptr_chunk_buf_len, 0)) == NULL) {
	free(entitize_buf);
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }

      if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* end page */
	free(entitize_buf);
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }
      free(entitize_buf);
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "</biblScope>\n", ptr_chunk_buf_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
  }

  /* reuse periodical buffer for date*/
  if ((item = get_refdb_pubyear(ptr_rendinfo->dbires, periodical)) != NULL) {
    if ((new_ref = mstrcat(*ptr_chunk_buf, "<date>", ptr_chunk_buf_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, periodical, ptr_chunk_buf_len, 0)) == NULL) { /* year */
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "</date>\n", ptr_chunk_buf_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
    n_have_data++;
  }

  /* publisher and city */
  if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "PB") != NULL
      || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
    if ((item = get_refdb_publisher(ptr_rendinfo->dbires)) != NULL) {
      if ((new_ref = mstrcat(*ptr_chunk_buf, "<publisher>", ptr_chunk_buf_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }

      if ((new_ref = mstrcat(*ptr_chunk_buf, (char*)item, ptr_chunk_buf_len, 0)) == NULL) { /* publisher */
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }
      
      if ((new_ref = mstrcat(*ptr_chunk_buf, "</publisher>", ptr_chunk_buf_len, 0)) == NULL) {
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }
      
      if (strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "CY") != NULL
	  || strstr((ptr_rendinfo->ptr_biblio_info)->format_string, "ALL") != NULL) {
	/* pubplace - use only if publisher is known */
	if ((item = get_refdb_city(ptr_rendinfo->dbires)) != NULL) {
	  if ((new_ref = mstrcat(*ptr_chunk_buf, "<pubplace>", ptr_chunk_buf_len, 0)) == NULL) {
	    LOG_PRINT(LOG_CRIT, get_status_msg(801));
	    return NULL;
	  }
	  else {
	    *ptr_chunk_buf = new_ref;
	  }

	  if ((new_ref = mstrcat(*ptr_chunk_buf, (char*)item, ptr_chunk_buf_len, 0)) == NULL) { /* city */
	    LOG_PRINT(LOG_CRIT, get_status_msg(801));
	    return NULL;
	  }
	  else {
	    *ptr_chunk_buf = new_ref;
	  }

	  if ((new_ref = mstrcat(*ptr_chunk_buf, "</pubplace>", ptr_chunk_buf_len, 0)) == NULL) {
	    LOG_PRINT(LOG_CRIT, get_status_msg(801));
	    return NULL;
	  }
	  else {
	    *ptr_chunk_buf = new_ref;
	  }
	}
      }
      
      n_have_data++;
    }
  }

  if ((new_ref = mstrcat(*ptr_chunk_buf, "</imprint>\n", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  /* finish monographic information */
  if ((new_ref = mstrcat(*ptr_chunk_buf, "</monogr>\n", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  if (!n_have_data) {
    **ptr_chunk_buf = '\0';
  }

  return *ptr_chunk_buf;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  add_series_teix() adds series level info to a TEI XML bibliography

  static char* add_series_teix returns ptr to ref if successful, NULL if not

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  int n_refdb_id ID of the current dataset

  char **ptr_chunk_buf ptr to an allocated string, may be resized

  size_t *ptr_chunk_buf_len ptr to length of ptr_chunk_buf

  const char* type ptr to string containing reference type

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static char* add_series_teix(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len, const char* type) {
  int n_have_data = 0; /* set to 1 if we have some data */
  char *new_ref;
  char *entitize_buf;
  char *author_buf;
  char *author_result;
  const char *item;
  size_t author_buf_len;

  **ptr_chunk_buf = '\0';

  author_buf_len = 1024;

  if ((author_buf = malloc(author_buf_len)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
    
  /* start series info */
  if ((new_ref = mstrcat(*ptr_chunk_buf, "<series>\n", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  /* add editors, if any */
  author_result = add_author_info_teix(ptr_rendinfo, 3, n_refdb_id, &author_buf, &author_buf_len);

  if (!author_result) {
    free(author_buf);
    return NULL;
  }
  else if (*author_result) {
    n_have_data++;
    if ((new_ref = mstrcat(*ptr_chunk_buf, author_result, ptr_chunk_buf_len, 0)) == NULL) {
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      free(author_buf);
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
  }

  free(author_buf);

  /* series title, if any */
  item = get_refdb_title_series_copy(ptr_rendinfo->dbires);

  if (item != NULL) {
    if ((entitize_buf = mstrdup((char*)item)) == NULL) {
      free((char*)item);
      return NULL;
    }

    if (sgml_entitize(&entitize_buf, NULL) == NULL) {
      free((char*)item);
      free(entitize_buf);
      return NULL;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "<title level=\"s\">", ptr_chunk_buf_len, 0)) == NULL) {
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* title */
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, "</title>\n", ptr_chunk_buf_len, 0)) == NULL) {
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
    free(entitize_buf);
    free((char*)item);
    n_have_data++;
  }

  /* finish series information */
  if ((new_ref = mstrcat(*ptr_chunk_buf, "</series>\n", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    free(author_buf);
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  if (!n_have_data) {
    **ptr_chunk_buf = '\0';
  }

  return *ptr_chunk_buf;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  add_author_info_teix() adds author info to a TEI XML bibliography

  static char* add_author_info_teix returns ptr to ref if successful, NULL if not

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  int author_type 1=author 2=editor 3=series editor

  int n_refdb_id ID of the current dataset

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static char* add_author_info_teix(struct renderinfo* ptr_rendinfo, int author_type, int n_refdb_id, char** ptr_chunk_buf, size_t* ptr_chunk_buf_len) {
  int n_have_data = 0;
  char *new_ref;
  char *authorstring;
  char *entitize_buf;
  char *item;
  char author_start[3][9] = {"<author>",
			     "<editor>",
			     "<editor>"};
  char author_end[3][10] = {"</author>",
			    "</editor>",
			    "</editor>"};
  dbi_conn conn;
  dbi_result dbires;
  struct AUTHOR_INFO ainfo;

  **ptr_chunk_buf = '\0';

  conn = dbi_result_get_conn(ptr_rendinfo->dbires);

  /* get authors */
  dbires = request_authors(conn, author_type, NULL, 0, n_refdb_id);
  if (dbires == NULL) {
    return NULL;
  }

  /* get buffer to assemble the docbook author output */
  authorstring = malloc((size_t)2048);  /* fixed length ok afaik */
    
  if (authorstring == NULL) {
    clean_request(dbires);
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }

  /* fetch all authors of this article */
  while ((get_author_parts(dbires, &ainfo)) != NULL) {
    authorstring[0] = '\0';

    if (!*(ainfo.lastname) && !*(ainfo.firstname)
      && !*(ainfo.middlename) && !*(ainfo.suffix)) {
      if (*(ainfo.name)) {
 	n_have_data++;
	strcat(authorstring, "<author><orgName>");
	strcat(authorstring, ainfo.name);
	strcat(authorstring, "</orgName></author>");
      }
    }
    else {
      strcat(authorstring, author_start[author_type-1]);
      strcat(authorstring, "<persName>");
      if (*(ainfo.firstname)) {
 	n_have_data++;
	strcat(authorstring, "<foreName>");
	if ((entitize_buf = mstrdup(ainfo.firstname)) == NULL) {
	  free(authorstring);
	  clean_request(dbires);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return NULL;
	}

	if (sgml_entitize(&entitize_buf, NULL) == NULL) {
	  free(entitize_buf);
	  free(authorstring);
	  clean_request(dbires);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return NULL;
	}

	strcat(authorstring, entitize_buf);
	free(entitize_buf);
	strcat(authorstring, "</foreName>");
      }

      if (*(ainfo.middlename)) {
 	n_have_data++;
	for (item = strtok(ainfo.middlename, " "); item; item = strtok(NULL, " ")) {
	  strcat(authorstring, "<addName>");
	  if ((entitize_buf = mstrdup(item)) == NULL) {
	    free(authorstring);
	    clean_request(dbires);
	    LOG_PRINT(LOG_CRIT, get_status_msg(801));
	    return NULL;
	  }

	  if (sgml_entitize(&entitize_buf, NULL) == NULL) {
	    free(entitize_buf);
	    free(authorstring);
	    clean_request(dbires);
	    LOG_PRINT(LOG_CRIT, get_status_msg(801));
	    return NULL;
	  }

	  strcat(authorstring, entitize_buf);
	  free(entitize_buf);
	  strcat(authorstring, "</addName>");
	}
      }

      if (*(ainfo.lastname)) {
 	n_have_data++;
	strcat(authorstring, "<surname>");
	if ((entitize_buf = mstrdup(ainfo.lastname)) == NULL) {
	  free(authorstring);
	  clean_request(dbires);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return NULL;
	}

	if (sgml_entitize(&entitize_buf, NULL) == NULL) {
	  free(entitize_buf);
	  free(authorstring);
	  clean_request(dbires);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return NULL;
	}

	strcat(authorstring, entitize_buf);
	free(entitize_buf);
	strcat(authorstring, "</surname>");
      }

      if (*(ainfo.suffix)) {
 	n_have_data++;
	strcat(authorstring, "<genName>");
	if ((entitize_buf = mstrdup(ainfo.suffix)) == NULL) {
	  free(authorstring);
	  clean_request(dbires);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return NULL;
	}

	if (sgml_entitize(&entitize_buf, NULL) == NULL) {
	  free(entitize_buf);
	  free(authorstring);
	  clean_request(dbires);
	  LOG_PRINT(LOG_CRIT, get_status_msg(801));
	  return NULL;
	}

	strcat(authorstring, entitize_buf);
	free(entitize_buf);
	strcat(authorstring, "</genName>");
      }
      strcat(authorstring, "</persName>");
      strcat(authorstring, author_end[author_type-1]);
    }
    if (n_have_data) {
      if ((new_ref = mstrcat(*ptr_chunk_buf, authorstring, ptr_chunk_buf_len, 0)) == NULL) {
	clean_request(dbires);
	free(authorstring);
	LOG_PRINT(LOG_CRIT, get_status_msg(801));
	return NULL;
      }
      else {
	*ptr_chunk_buf = new_ref;
      }
    }
  }
  free(authorstring);
  clean_request(dbires);

  /* if there was no author info, chunk_buf is returned unaltered */
  return *ptr_chunk_buf;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  add_notes_teix() adds notes info to a TEI bibliography

  static char* add_notes_teix returns ptr to ref if successful, NULL if not

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  int n_refdb_id ID of the current dataset

  char **ptr_chunk_buf ptr to an allocated string, may be resized

  size_t *ptr_chunk_buf_len ptr to length of ptr_chunk_buf

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static char* add_notes_teix(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len) {
  int n_have_data = 0; /* set to 1 if we have some data */
  char *new_ref;
  char *entitize_buf;
  const char *item;

  **ptr_chunk_buf = '\0';

  /* start bibliomisc element */
  if ((new_ref = mstrcat(*ptr_chunk_buf, "<note type=\"annotation\">", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  /* note of current user, if any */
  item = get_notes_copy(ptr_rendinfo->dbires, ptr_rendinfo->username);

  if (item && *item) {
    n_have_data++;
    if ((entitize_buf = mstrdup((char*)item)) == NULL) {
      free((char*)item);
      return NULL;
    }

    if (sgml_entitize(&entitize_buf, NULL) == NULL) {
      free((char*)item);
      free(entitize_buf);
      return NULL;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* title */
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }
  }

  /* finish notes */
  if ((new_ref = mstrcat(*ptr_chunk_buf, "</note>\n", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  if (!n_have_data) {
    **ptr_chunk_buf = '\0';
  }

  return *ptr_chunk_buf;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  add_abstract_teix() adds abstract to a TEI bibliography

  static char* add_abstract_db returns ptr to ref if successful, NULL if not

  struct renderinfo* ptr_rendinfo ptr to a structure with the info
                             how the reference should be rendered

  int n_refdb_id ID of the current dataset

  char **ptr_chunk_buf ptr to an allocated string, may be resized

  size_t *ptr_chunk_buf_len ptr to length of ptr_chunk_buf

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static char* add_abstract_teix(struct renderinfo* ptr_rendinfo, int n_refdb_id, char **ptr_chunk_buf, size_t *ptr_chunk_buf_len) {
  int n_have_data = 0; /* set to 1 if we have some data */
  char *new_ref;
  char *entitize_buf;
  const char *item;

  **ptr_chunk_buf = '\0';

  /* start abstract */
  if ((new_ref = mstrcat(*ptr_chunk_buf, "<note type=\"abstract\">", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  /* abstract, if any */
  item = get_refdb_abstract_copy(ptr_rendinfo->dbires);

  if (item != NULL) {
    n_have_data++;
    if ((entitize_buf = mstrdup((char*)item)) == NULL) {
      free((char*)item);
      return NULL;
    }

    if (sgml_entitize(&entitize_buf, NULL) == NULL) {
      free((char*)item);
      free(entitize_buf);
      return NULL;
    }

    if ((new_ref = mstrcat(*ptr_chunk_buf, entitize_buf, ptr_chunk_buf_len, 0)) == NULL) { /* abstract */
      free((char*)item);
      free(entitize_buf);
      LOG_PRINT(LOG_CRIT, get_status_msg(801));
      return NULL;
    }
    else {
      *ptr_chunk_buf = new_ref;
    }

    free(entitize_buf);
    free((char*)item);
  }

  /* finish abstract information */
  if ((new_ref = mstrcat(*ptr_chunk_buf, "</abstract>\n", ptr_chunk_buf_len, 0)) == NULL) {
    LOG_PRINT(LOG_CRIT, get_status_msg(801));
    return NULL;
  }
  else {
    *ptr_chunk_buf = new_ref;
  }

  if (!n_have_data) {
    **ptr_chunk_buf = '\0';
  }

  return *ptr_chunk_buf;
}
